package org.expeditee.actions;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.GraphicsEnvironment;
import java.awt.Image;
import java.awt.image.BufferedImage;
import java.awt.image.VolatileImage;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.List;

import javax.imageio.ImageIO;

import org.expeditee.gui.DisplayIO;
import org.expeditee.gui.Frame;
import org.expeditee.gui.FrameGraphics;
import org.expeditee.gui.FrameIO;
import org.expeditee.gui.FrameMouseActions;
import org.expeditee.gui.TimeKeeper;
import org.expeditee.items.Item;
import org.expeditee.items.Text;
import org.expeditee.stats.SessionStats;
import org.expeditee.stats.StatsLogger;

/**
 * A list of miscellaneous KMS Actions and Actions specific to Expeditee
 * 
 */
public class Misc {
	private static final int FRAME_FILE_FONT_SIZE = 10;

	/**
	 * Causes the system to beep
	 */
	public static void Beep() {
		java.awt.Toolkit.getDefaultToolkit().beep();
	}

	/**
	 * Forces a repaint of the current Frame
	 */
	public static void Display() {
		FrameGraphics.ForceRepaint();
	}

	/**
	 * Restores the current frame to the last saved version currently on the
	 * hard disk
	 */
	public static void Restore() {
		FrameIO.Reload();

		FrameGraphics.DisplayMessage("Restoration complete.");
	}

	/**
	 * Toggles AudienceMode on or off
	 */
	public static void ToggleAudienceMode() {
		FrameGraphics.ToggleAudienceMode();
	}

	/**
	 * Toggles TwinFrames mode on or off
	 */
	public static void ToggleTwinFramesMode() {
		DisplayIO.ToggleTwinFrames();
	}

	/**
	 * Exits the System
	 */
	public static void Exit() {
		/**
		 * TODO: Prompt the user etc.
		 */
		FrameIO.SaveFrame(DisplayIO.getCurrentFrame());

		while (FrameIO.DeleteFrameset("messages"))
			;

		StatsLogger.WriteStatsFile();

		System.exit(0);
	}

	/**
	 * If the given Item is a Text Item, then the text of the Item is
	 * interpreted as actions, if not this method does nothing.
	 * 
	 * @param current
	 *            The Item to read the Actions from
	 */
	public static void RunCurrentItem(Item current) {
		if (current instanceof Text) {
			List<String> actions = ((Text) current).getTextList();

			for (String action : actions) {
				Actions.PerformAction(DisplayIO.getCurrentFrame(), current,
						action);
			}
		}

	}

	/**
	 * Prompts the user to confirm deletion of the current Frame, and deletes if
	 * the user chooses. After deletion this action calls back(), to ensure the
	 * deleted frame is not still being shown
	 * 
	 */
	public static void DeleteFrame() {
		Frame toDelete = DisplayIO.getCurrentFrame();
		DisplayIO.Back();
		String deletedFrame = toDelete.getName();
		String deletedFrameNameLowercase = deletedFrame.toLowerCase();
		try {
			boolean del = FrameIO.DeleteFrame(toDelete);
			if (!del) {
				FrameGraphics.ErrorMessage("Error trying to delete "
						+ toDelete.getName());
			} else {
				Frame current = DisplayIO.getCurrentFrame();
				for (Item i : current.getItems())
					if (i.getLink() != null
							&& i.getAbsoluteLink().toLowerCase().equals(
									deletedFrameNameLowercase)) {
						i.setLink(null);
					}

				FrameGraphics.Repaint();
				FrameGraphics.DisplayMessage(deletedFrame + " Deleted.");
			}
		} catch (IOException ioe) {
			FrameGraphics.ErrorMessage("Error trying to delete " + deletedFrame
					+ ":\n" + ioe.getMessage());
		}
	}

	/**
	 * Loads the Frame linked to by the given Item. The first Item on the Frame
	 * that is not the title or name is then placed on the cursor. If the given
	 * Item has no link, or no item is found then this is a no-op.
	 * 
	 * @param current
	 *            The Item that links to the Frame that the Item will be loaded
	 *            from.
	 */
	public static void GetItemFromChildFrame(Item current) {
		getFromChildFrame(current, false);
	}

	/**
	 * Loads the Frame linked to by the given Item. The first Text Item on the
	 * Frame that is not the title or name is then placed on the cursor. If the
	 * given Item has no link, or no item is found then this is a no-op.
	 * 
	 * @param current
	 *            The Item that links to the Frame that the Item will be loaded
	 *            from.
	 */
	public static void GetTextFromChildFrame(Item current) {
		getFromChildFrame(current, true);
	}

	private static void getFromChildFrame(Item current, boolean textOnly) {
		Item item = getFirstBodyItemOnChildFrame(current, textOnly);
		// if no item was found
		if (item == null) {
			return;
		}

		// copy the item and switch
		item = item.copy();
		item.setPosition(DisplayIO.getMouseX(), DisplayIO.getMouseY());
		item.setParent(null);

		FrameMouseActions.pickup(item);
		FrameGraphics.Repaint();
	}

	/**
	 * Sets the given Item to have the Given Color. Color can be null (for
	 * default)
	 * 
	 * @param toChange
	 *            The Item to set the Color.
	 * @param toUse
	 *            The Color to give the Item.
	 */
	public static void SetCurrentItemBackgroundColor(Item toChange, Color toUse) {
		if (toChange == null)
			return;

		toChange.setBackgroundColor(toUse);
		FrameGraphics.Repaint();
	}

	/**
	 * Sets the given Item to have the Given Color. Color can be null (for
	 * default)
	 * 
	 * @param toChange
	 *            The Item to set the Color.
	 * @param toUse
	 *            The Color to give the Item.
	 */
	public static void SetCurrentItemColor(Item toChange, Color toUse) {
		if (toChange == null)
			return;

		toChange.setColor(toUse);
		FrameGraphics.Repaint();
	}

	/**
	 * Creates a new Text Object containing general statistics for the current
	 * session. The newly created Text Object is then attached to the cursor via
	 * FrameMouseActions.pickup(Item)
	 */
	public static void GetSessionStats() {
		CreateTextItem(SessionStats.getCurrentStats());
	}

	public static void CreateTextItem(String itemText) {
		SessionStats.CreatedText();
		Frame current = DisplayIO.getCurrentFrame();
		Item text = current.getStatsTextItem(itemText);
		FrameMouseActions.pickup(text);
		FrameGraphics.Repaint();
	}

	/**
	 * Creates a new Text Object containing statistics for moving, deleting and
	 * creating items in the current session. The newly created Text Object is
	 * then attached to the cursor via FrameMouseActions.pickup(Item)
	 */
	public static void GetItemStats() {
		CreateTextItem(SessionStats.getItemStats());
	}

	/**
	 * Creates a new Text Object containing statistics for the time between
	 * events triggered by the user through mouse clicks and key presses. The
	 * newly created Text Object is then attached to the cursor via
	 * FrameMouseActions.pickup(Item)
	 */
	public static void GetEventStats() {
		CreateTextItem(SessionStats.getEventStats());
	}

	/**
	 * Creates a new Text Object containing the contents of the current frames
	 * file.
	 */
	public static void GetCurrentFrameFile() {

		String fileContents = FrameIO.SaveFrame(DisplayIO.getCurrentFrame());

		Text text = DisplayIO.getCurrentFrame().createNewText();
		// We dont want the stats to wrap at all
		text.setMaxSize(new Dimension(Integer.MAX_VALUE, Integer.MAX_VALUE));

		text.setText(fileContents);

		text.setPosition(DisplayIO.getMouseX(), DisplayIO.getMouseY());
		text.setSize(FRAME_FILE_FONT_SIZE);

		FrameMouseActions.pickup(text);

		FrameGraphics.Repaint();
	}

	/**
	 * Creates a new Text Object containing the available fonts.
	 */
	public static void GetAvailableFontFamilyNames() {

		String[] availableFonts = GraphicsEnvironment
				.getLocalGraphicsEnvironment().getAvailableFontFamilyNames();
		StringBuilder fontsList = new StringBuilder();
		for (String s : availableFonts) {
			fontsList.append(s).append('\n');
		}

		Text text = DisplayIO.getCurrentFrame().createNewText();
		// We dont want the stats to wrap at all
		text.setMaxSize(new Dimension(Integer.MAX_VALUE, Integer.MAX_VALUE));
		text.setText(fontsList.toString());

		text.setPosition(DisplayIO.getMouseX(), DisplayIO.getMouseY());
		text.setSize(FRAME_FILE_FONT_SIZE);

		FrameMouseActions.pickup(text);

		FrameGraphics.Repaint();
	}

	/**
	 * Resets the statistics back to zero.
	 */
	public static void ResetStats() {
		StatsLogger.WriteStatsFile();
		SessionStats.resetStats();
	}

	/**
	 * Loads a frame with the given name and saves it as a JPEG image.
	 * 
	 * @param framename
	 *            The name of the Frame to save
	 */
	public static void JpegFrame(String framename) {
		ImageFrame(framename, "JPEG");
	}

	/**
	 * Saves the current frame as a JPEG image. This is the same as calling
	 * JpegFrame(currentFrame.getName())
	 */
	public static void JpegFrame() {
		ImageFrame(DisplayIO.getCurrentFrame().getName(), "JPEG");
	}

	public static void JPGFrame() {
		JpegFrame();
	}

	/**
	 * Loads a frame with the given name and saves it as a PNG image.
	 * 
	 * @param framename
	 *            The name of the Frame to save
	 */
	public static void PNGFrame(String framename) {
		ImageFrame(framename, "PNG");
	}

	/**
	 * Saves the current frame as a PNG image. This is the same as calling
	 * PNGFrame(currentFrame.getName())
	 */
	public static void PNGFrame() {
		ImageFrame(DisplayIO.getCurrentFrame().getName(), "PNG");
	}

	public static String SaveImage(BufferedImage screen, String format, String directory, String fileName){
		// Check if we need to append the suffix
		if (fileName.indexOf('.') < 0)
			fileName += "." + format.toLowerCase();		

		try {
			// set up the file for output
			String fullFileName = directory + fileName;
			File out = new File(fullFileName);
			if (!out.getParentFile().exists())
				out.mkdirs();

			// If the image is successfully written out return the fileName
			if (ImageIO.write(screen, format, out))
				return fileName;

		} catch (Exception e) {
			e.printStackTrace();
		}
		return null;
	}
	
	public static String ImageFrame(Frame frame, String format, String directory) {
		assert (frame != null);
		
		FrameGraphics.UpdateBuffer(frame, false);

		BufferedImage screen = null;
		Image frameBuffer = frame.getBuffer();
		if (frame.getBuffer() instanceof BufferedImage) {
			screen = (BufferedImage) frameBuffer;
		} else if (frameBuffer instanceof VolatileImage) {
			screen = ((VolatileImage) frameBuffer).getSnapshot();
		} else {
			assert (false);
		}
		return SaveImage(screen, format, directory, frame.getExportFileName());
	}

	/**
	 * Saves the Frame with the given Framename as an image of the given format.
	 * 
	 * @param framename
	 *            The name of the Frame to save as an image
	 * @param format
	 *            The Image format to use (i.e. "PNG", "BMP", etc)
	 */
	public static void ImageFrame(String framename, String format) {
		Frame loaded = FrameIO.LoadFrame(framename);

		// if the frame was loaded successfully
		if (loaded != null) {
			String path = FrameIO.IMAGES_PATH;
			String frameName = ImageFrame(loaded, format, path);
			if (frameName != null)
				FrameGraphics.DisplayMessage("Frame successfully saved to "
						+ path + frameName);
			else
				FrameGraphics.ErrorMessage("Could not find image writer for "
						+ format + " format");
			// if the frame was not loaded successfully, alert the user
		} else
			FrameGraphics.DisplayMessage("Frame '" + framename
					+ "' could not be found.");
	}

	/**
	 * Displays a message in the message box area.
	 * 
	 * @param message
	 *            the message to display
	 */
	public static void MessageLn(String message) {
		FrameGraphics.DisplayMessage(message);
	}

	public static void MessageLn2(String message, String message2) {
		FrameGraphics.DisplayMessage(message + " " + message2);
	}

	public static void CopyFile(String existingFile, String newFileName) {
		try {
			// TODO is there a built in method which will do this faster?

			FrameGraphics.DisplayMessage("Copying file " + existingFile
					+ " to " + newFileName + "...");
			FrameIO.copyFile(existingFile, newFileName);
			FrameGraphics.DisplayMessage("File copied successfully");
		} catch (FileNotFoundException e) {
			FrameGraphics.DisplayMessage("Error opening file: " + existingFile);
		} catch (Exception e) {
			FrameGraphics.DisplayMessage("File could not be copied");
		}
	}

	/**
	 * Runs two methods alternatively a specified number of times and reports on
	 * the time spent running each method.
	 * 
	 * @param fullMethodNameA
	 * @param fullMethodNameB
	 * @param repsPerTest
	 *            the number of time each method is run per test
	 * @param tests
	 *            the number of tests to conduct
	 * 
	 */
	public static void CompareMethods(String fullMethodNameA,
			String fullMethodNameB, int repsPerTest, int tests) {
		try {
			String classNameA = getClassName(fullMethodNameA);
			String classNameB = getClassName(fullMethodNameB);
			String methodNameA = getMethodName(fullMethodNameA);
			String methodNameB = getMethodName(fullMethodNameB);

			Class classA = Class.forName(classNameA);
			Class classB = Class.forName(classNameB);
			Method methodA = classA.getDeclaredMethod(methodNameA,
					new Class[] {});
			Method methodB = classB.getDeclaredMethod(methodNameB,
					new Class[] {});
			TimeKeeper timeKeeper = new TimeKeeper();
			long timeA = 0;
			long timeB = 0;
			// Run the tests
			for (int i = 0; i < tests; i++) {
				// Test methodA
				timeKeeper.restart();
				for (int j = 0; j < repsPerTest; j++) {
					methodA.invoke((Object) null, new Object[] {});
				}
				timeA += timeKeeper.getElapsedMillis();
				timeKeeper.restart();
				// Test methodB
				for (int j = 0; j < repsPerTest; j++) {
					methodB.invoke((Object) null, new Object[] {});
				}
				timeB += timeKeeper.getElapsedMillis();
			}

			float aveTimeA = timeA * 1000F / repsPerTest / tests;
			float aveTimeB = timeB * 1000F / repsPerTest / tests;
			// Display Results
			FrameGraphics.DisplayMessage("Average Execution Time");
			FrameGraphics.DisplayMessage(methodNameA + ": "
					+ TimeKeeper.Formatter.format(aveTimeA) + "us");
			FrameGraphics.DisplayMessage(methodNameB + ": "
					+ TimeKeeper.Formatter.format(aveTimeB) + "us");
		} catch (Exception e) {
			FrameGraphics.ErrorMessage(e.getClass().getSimpleName() + ": "
					+ e.getMessage());
		}
	}

	public static String getClassName(String fullMethodName) {
		assert (fullMethodName != null);
		assert (fullMethodName.length() > 0);
		int lastPeriod = fullMethodName.lastIndexOf('.');
		if (lastPeriod > 0 && lastPeriod < fullMethodName.length() - 1)
			return fullMethodName.substring(0, lastPeriod);
		throw new RuntimeException("Invalid method name: " + fullMethodName);
	}

	public static String getMethodName(String methodName) {
		assert (methodName != null);
		assert (methodName.length() > 0);
		int lastPeriod = methodName.lastIndexOf('.');
		if (lastPeriod > 0 && lastPeriod < methodName.length() - 1)
			return methodName.substring(1 + lastPeriod);
		throw new RuntimeException("Invalid method name: " + methodName);
	}

	/**
	 * Loads the Frame linked to by the given Item. The first Item on the Frame
	 * that is not the title or name is then placed on the current frame. The
	 * item that was clicked on is placed on the frame it was linked to and the
	 * link is switched to the item from the child frame. If the given Item has
	 * no link, or no item is found then this is a no-op.
	 * 
	 * @param current
	 *            The Item that links to the Frame that the Item will be loaded
	 *            from.
	 */
	public static void SwapItemWithItemOnChildFrame(Item current) {
		Item item = getFirstBodyItemOnChildFrame(current, false);
		// if no item was found
		if (item == null) {
			return;
		}

		// swap the items parents
		Frame parentFrame = current.getParent();
		Frame childFrame = item.getParent();
		current.setParent(childFrame);
		item.setParent(parentFrame);

		// swap the items on the frames
		parentFrame.removeItem(current);
		childFrame.removeItem(item);
		parentFrame.addItem(item);
		childFrame.addItem(current);

		// swap the items links
		item.setActions(current.getAction());
		item.setLink(childFrame.getName());
		current.setLink(parentFrame.getName());
		// current.setLink(null);
		current.setActions(null);

		FrameGraphics.Repaint();
	}

	private static Item getFirstBodyItemOnChildFrame(Item current,
			boolean textOnly) {
		// the item must link to a frame
		if (current.getLink() == null) {
			FrameGraphics
					.DisplayMessage("Cannot get item from child - this item has no link");
			return null;
		}

		Frame child = FrameIO.LoadFrame(current.getAbsoluteLink());

		// if the frame could not be loaded
		if (child == null) {
			FrameGraphics.ErrorMessage("Could not load child frame.");
			return null;
		}

		// find the first non-title and non-name item
		List<Item> body = new ArrayList<Item>();
		if (textOnly)
			body.addAll(child.getBodyTextItems(false));
		else
			body.addAll(child.getItems());
		Item item = null;

		for (Item i : body)
			if (i != child.getTitleItem() && !i.isAnnotation()) {
				item = i;
				break;
			}

		// if no item was found
		if (item == null) {
			FrameGraphics.DisplayMessage("No item found to copy");
			return null;
		}

		return item;
	}
}
