package org.expeditee.gui;

import java.awt.Color;
import java.awt.Polygon;
import java.io.File;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

import org.expeditee.io.Logger;
import org.expeditee.items.Dot;
import org.expeditee.items.InteractiveWidget;
import org.expeditee.items.InteractiveWidgetNotAvailableException;
import org.expeditee.items.Item;
import org.expeditee.items.ItemUtils;
import org.expeditee.items.Line;
import org.expeditee.items.Picture;
import org.expeditee.items.Text;
import org.expeditee.items.WidgetCorner;
import org.expeditee.items.WidgetEdge;
import org.expeditee.stats.SessionStats;

public class FrameUtils {

	private static final int COLUMN_WIDTH = 50;

	/**
	 * Provides a way to monitor the time elapsed between button-down and the
	 * finished painting.
	 */
	public static TimeKeeper ResponseTimer = new TimeKeeper();

	private static float _ResponseTimeSum = 0;

	private static float _LastResponse = 0;

	public static Item LastEdited = null;

	public static int MINIMUM_INTERITEM_SPACING = -6;

	public static float getResponseTimeTotal() {
		return _ResponseTimeSum;
	}

	public static float getLastResponseTime() {
		return _LastResponse;
	}

	/**
	 * Checks if the given top Item is above the given bottom Item, allowing for
	 * the X coordinates to be off by a certain width...
	 * 
	 * @param item1
	 *            The Item to check is above the other Item
	 * @param item2
	 *            The Item to check is below the top Item
	 * @return True if top is above bottom, False otherwise.
	 */
	public static boolean inSameColumn(Item item1, Item item2) {
		if (!(item1 instanceof Text) || !(item2 instanceof Text))
			return false;

		if (item1.getID() < 0 || item2.getID() < 0)
			return false;

		int minX = item2.getX();
		int maxX = item2.getX() + item2.getBoundsWidth();

		int startX = item1.getX();
		int endX = item1.getX() + item1.getBoundsWidth();

		// Check that the two items left values are close
		if (Math.abs(item1.getX() - item2.getX()) > COLUMN_WIDTH)
			return false;

		// Ensure the two items
		if ((minX >= startX && minX <= endX)
				|| (maxX >= startX && maxX <= endX)
				|| (startX >= minX && startX <= maxX)
				|| (endX >= minX && endX <= maxX))
			return true;

		return false;
	}

	public static boolean sameBulletType(String bullet1, String bullet2) {
		if (bullet1 == null || bullet2 == null)
			return false;

		if (bullet1.equals("") || bullet2.equals(""))
			return false;

		if (Character.isLetter(bullet1.charAt(0))
				&& Character.isLetter(bullet2.charAt(0)))
			return true;

		if (Character.isDigit(bullet1.charAt(0))
				&& Character.isDigit(bullet2.charAt(0)))
			return true;

		// TODO make this more sofisticated

		return false;
	}

	private static boolean needsRenumbering(String s) {
		if (s == null || s.equals(""))
			return false;
		if (!Character.isLetterOrDigit(s.charAt(0)))
			return false;

		s = s.trim();
		// if its all letters then we dont want to auto adjust
		if (s.length() > 2) {
			for (int i = 0; i < s.length() - 1; i++) {
				if (!Character.isLetter(s.charAt(i)))
					return true;
			}
		} else
			return true;

		return false;
	}

	/**
	 * 
	 * @param toAlign
	 * @param moveAll
	 * @param adjust
	 * @return
	 */
	public static int Align(List<Item> toAlign, boolean moveAll, int adjust) {
		Collections.sort(toAlign);

		// Single items dont need alignment
		if (toAlign.size() < 2)
			return 0;

		// get the first item
		Item from = toAlign.get(0);
		if (from.getParent() == null)
			from = toAlign.get(1);
		int x = from.getX();

		Frame curr = from.getParent();
		Item above = curr.getItemAbove(from);

		String lastBullet = "";

		if (above != null && curr.isNormalTextItem(above))
			lastBullet = FrameKeyboardActions.getAutoBullet(((Text) above)
					.getText());
		else {
			lastBullet = FrameKeyboardActions.getBullet(((Text) toAlign.get(0))
					.getText());
		}
		if (needsRenumbering(lastBullet)) {
			// renumber...
			for (int i = 0; i < toAlign.size(); i++) {
				assert toAlign.get(i) instanceof Text;

				Text currentText = ((Text) toAlign.get(i));
				String currentBullet = FrameKeyboardActions
						.getAutoBullet(currentText.getText());

				if (sameBulletType(lastBullet, currentBullet)) {
					currentText.stripFirstWord();

					currentText.setText(lastBullet + currentText.getText());
					lastBullet = FrameKeyboardActions.getAutoBullet(currentText
							.getText());
				}
			}
		}

		// work out the spacing between the first item and the one above it

		int space = 10 + adjust;

		// if we are dropping from the title make the space a little bigger
		// than normal
		if (above != curr.getTitleItem() && above != null) {
			// Make the gap between all items the same as the gap between
			// the first two
			space = (int) (from.getPolygon().getBounds().getMinY() - above
					.getPolygon().getBounds().getMaxY());

			if (space < MINIMUM_INTERITEM_SPACING)
				space = MINIMUM_INTERITEM_SPACING;

			if (above != curr.getNameItem() && above != curr.getTitleItem())
				x = above.getX();

			space += adjust;

			from.setY((int) above.getPolygon().getBounds().getMaxY()
					+ space
					+ ((int) (from.getY() - from.getPolygon().getBounds()
							.getMinY())));

			if (moveAll)
				from.setX(x);
		}

		for (int i = 1; i < toAlign.size(); i++) {
			Item current = toAlign.get(i);
			Item top = toAlign.get(i - 1);

			// The bottom of the previous item
			int bottom = (int) top.getPolygon().getBounds().getMaxY();

			// the difference between the current item's Y coordinate and
			// the top of the highlight box
			int diff = (int) (current.getY() - current.getPolygon().getBounds()
					.getMinY());

			int newPos = bottom + space + diff;

			if (moveAll) {
				current.setPosition(x, newPos);
			} else if (newPos > current.getY()) {
				current.setY(newPos);
			}

		}

		// if (insert != null)
		// return insert.getY();

		// Michael thinks we return the y value for the next new item??
		int y = from.getY() + from.getBoundsHeight() + space;
		return y;
	}

	public static boolean LeavingFrame(Frame current) {
		checkTDFCItemWaiting(current);
		// active overlay frames may also require saving if they have been
		// changed
		for (Overlay o : current.getOverlays())
			if (!SaveCheck(o.Frame))
				return false;

		// if the check fails there is no point continuing
		if (!SaveCheck(current))
			return false;

		for (Item i : current.getItems())
			i.setSelectedMode(Item.SelectedMode.None);
		return true;
	}

	private static boolean SaveCheck(Frame toSave) {
		// don't bother saving frames that haven't changed
		if (!toSave.hasChanged())
			return true;

		// if the frame has been changed, then save it
		if (DisplayIO.isTwinFramesOn()) {
			Frame opposite = DisplayIO.getOppositeFrame();

			String side = "left";
			if (DisplayIO.getCurrentSide() == 0)
				side = "right";

			// if the two frames both have changes, prompt the user for the
			// next
			// move
			if (opposite.hasChanged() && opposite.equals(toSave)) {
				if (DisplayIO.DisplayConfirmDialog(
						"Leaving this frame will discard changes made in the "
								+ side + " Frame. Continue?", "Changes",
						DisplayIO.TYPE_WARNING, DisplayIO.OPTIONS_OK_CANCEL,
						DisplayIO.RESULT_OK)) {
					FrameIO.SaveFrame(toSave);
					DisplayIO.Reload(DisplayIO.FrameOnSide(opposite));
					return true;
				} else
					return false;
			} else if (opposite.getOverlays().contains(new Overlay(toSave, 1))) {
				if (opposite.getOverlays().get(
						opposite.getOverlays().indexOf(new Overlay(toSave, 1))).Frame
						.hasChanged())
					if (DisplayIO.DisplayConfirmDialog(
							"Leaving this frame will discard changes made in the "
									+ side + " Frame. Continue?", "Changes",
							DisplayIO.TYPE_WARNING,
							DisplayIO.OPTIONS_OK_CANCEL, DisplayIO.RESULT_OK)) {
						FrameIO.SaveFrame(toSave);
						DisplayIO.Reload(DisplayIO.FrameOnSide(opposite));
						return true;
					} else
						return false;
			}

			// save the current frame and restore the other side
			FrameIO.SaveFrame(toSave);
			return true;
		}

		// single-frame mode can just save and return
		FrameIO.SaveFrame(toSave);
		return true;
	}

	/**
	 * Displays the given Frame on the display. If the current frame has changed
	 * since the last save then it will be saved before the switch is made. The
	 * caller can also dictate whether the current frame is added to the
	 * back-stack or not.
	 * 
	 * @param toDisplay
	 *            The Frame to display on the screen
	 * @param addToBack
	 *            True if the current Frame should be added to the back-stack,
	 *            False otherwise
	 */
	public static void DisplayFrame(Frame toDisplay, boolean addToBack) {
		if (toDisplay == null)
			return;

		Frame current = DisplayIO.getCurrentFrame();

		// move any anchored connected items
		if (Frame.itemAttachedToCursor()) {
			List<Item> toAdd = new ArrayList<Item>();
			List<Item> toCheck = new ArrayList<Item>(Frame.FreeItems);

			while (toCheck.size() > 0) {
				Item i = toCheck.get(0);
				Collection<Item> connected = i.getAllConnected();

				// // Only move completely enclosed items
				// if (!toCheck.containsAll(connected)) {
				// connected.retainAll(Frame.FreeItems);
				// Frame.FreeItems.removeAll(connected);
				// toCheck.removeAll(connected);
				// FrameMouseActions.anchor(connected);
				// } else {
				// toCheck.removeAll(connected);
				// }

				// Anchor overlay items where they belong
				if (i.getParent() != null && i.getParent() != current) {
					Frame.FreeItems.removeAll(connected);
					toCheck.removeAll(connected);
					FrameMouseActions.anchor(connected);
				} else {
					// Add stuff that is partially enclosed
					// remove all the connected items from our list to check
					toCheck.removeAll(connected);
					// Dont add the items that are free
					connected.removeAll(Frame.FreeItems);
					toAdd.addAll(connected);
				}
			}

			current.removeAllItems(toAdd);

			boolean oldChange = toDisplay.hasChanged();
			toDisplay.updateIDs(toAdd);
			toDisplay.addAllItems(toAdd);
			toDisplay.setChanged(oldChange);
		}

		// if the saving happened properly, we can continue
		if (!LeavingFrame(current)) {
			FrameGraphics.DisplayMessage("Navigation cancelled");
			return;
		}

		if (addToBack && current != toDisplay)
			DisplayIO.addToBack(current);

		Parse(toDisplay);
		DisplayIO.setCurrentFrame(toDisplay);

		// update response timer
		_LastResponse = ResponseTimer.getElapsedSeconds();
		_ResponseTimeSum += _LastResponse;
		DisplayIO.UpdateTitle();
	}

	/**
	 * Loads and displays the Frame with the given framename, and adds the
	 * current frame to the back-stack if required.
	 * 
	 * @param framename
	 *            The name of the Frame to load and display
	 * @param addToBack
	 *            True if the current Frame should be added to the back-stack,
	 *            false otherwise
	 */
	public static void DisplayFrame(String frameName, boolean addToBack) {
		Frame newFrame = getFrame(frameName);

		if (newFrame != null)
			// display the frame
			DisplayFrame(newFrame, addToBack);
	}

	/**
	 * Loads and displays the Frame with the given framename and adds the
	 * current frame to the back-stack. This is the same as calling
	 * DisplayFrame(framename, true)
	 * 
	 * @param framename
	 *            The name of the Frame to load and display
	 */
	public static void DisplayFrame(String framename) {
		DisplayFrame(framename, true);
	}

	public static void DisplayHomeFrame() {
		DisplayFrame(UserSettings.FirstFrame);
	}

	public static void DisplayProfileFrame() {
		DisplayFrame(UserSettings.Username + '1');
	}

	public static Frame getFrame(String frameName) {
		// if the new frame does not exist then tell the user
		Frame f = FrameIO.LoadFrame(frameName);

		if (f == null) {
			FrameGraphics.ErrorMessage("Frame '" + frameName
					+ "' could not be found.");
		}

		return f;
	}

	/**
	 * Creates a new Picture Item from the given Text source Item and adds it to
	 * the given Frame.
	 * 
	 * @return True if the image was created successfully, false otherwise
	 */
	private static boolean createPicture(Frame frame, Text txt) {
		// attempt to create the picture
		Picture pic = ItemUtils.CreatePicture(txt, frame);

		// if the picture could not be created successfully
		if (pic == null) {
			assert (txt.getFirstLine() != null);
			assert (txt.getFirstLine().split(" ").length > 1);

			FrameGraphics
					.ErrorMessage("Image " + txt.getFirstLine().split(" ")[1]
							+ " could not be loaded");
			return false;
		}

		// replace the Text source with the new Picture
		frame.removeItem(txt);
		frame.addItem(pic);

		return true;
	}

	private static boolean createFramePicture(Frame frame, Text txt) {
		if (txt.getLink() == null)
			return false;

		// attempt to create the picture
		Picture pic = ItemUtils.CreateFrameImage(txt, frame);

		// if the picture could not be created successfully
		if (pic == null) {
			FrameGraphics.ErrorMessage("Error loading frame image");
			return false;
		}

		// replace the Text source with the new Picture
		frame.removeItem(txt);
		frame.addItem(pic);

		return true;
	}

	private static boolean createFrameBitmap(Frame frame, Text txt) {
		if (txt.getLink() == null)
			return false;

		// attempt to create the picture
		Picture pic = ItemUtils.CreateFrameBitmap(txt, frame);

		// if the picture could not be created successfully
		if (pic == null) {
			FrameGraphics.ErrorMessage("Error loading frame image");
			return false;
		}

		// replace the Text source with the new Picture
		frame.removeItem(txt);
		frame.addItem(pic);

		return true;
	}

	/**
	 * Creates an interactive widget and adds it to a frame. If txt has no
	 * parent the parent will be set to frame.
	 * 
	 * @param frame
	 *            Frame to add widget to. Must not be null.
	 * 
	 * @param txt
	 *            Text to create the widget from. Must not be null.
	 * 
	 * @return True if created/added. False if coul not create.
	 * 
	 * @author Brook Novak
	 */
	private static boolean createWidget(Frame frame, Text txt) {

		if (frame == null)
			throw new NullPointerException("frame");
		if (txt == null)
			throw new NullPointerException("txt");

		// Safety
		if (txt.getParent() == null)
			txt.setParent(frame);

		InteractiveWidget iw = null;

		try {

			iw = InteractiveWidget.CreateWidget(txt);

		} catch (InteractiveWidgetNotAvailableException e) {
			e.printStackTrace();
			FrameGraphics.ErrorMessage("Cannot create iWidget: "
					+ e.getMessage());

		} catch (IllegalArgumentException e) {
			e.printStackTrace();
			FrameGraphics.ErrorMessage("Cannot create iWidget: "
					+ e.getMessage());
		}

		if (iw == null)
			return false;

		frame.removeItem(txt);

		frame.addAllItems(iw.getItems());

		return true;
	}

	public static void ParseProfile(Frame profile) {
		if (profile == null)
			return;

		UserSettings.TitleTemplate = profile.getTitleItem();

		List<Item> items = profile.getItems();

		// check for all tags setting user values
		for (Item item : items) {
			if (ItemUtils.startsWithTag(item, "@HomeFrame")) {
				String first = getLink(item, UserSettings.FirstFrame);
				// do not use non-existant frames as the first frame
				if (FrameIO.isValidFrameName(first)) {
					UserSettings.FirstFrame = first;
				}
				// warn the user
				else {
					FrameGraphics.WarningMessage("Home frame: " + first
							+ " is not a valid frame.");
					UserSettings.FirstFrame = profile.getName();
				}
			} else if (ItemUtils.startsWithTag(item, "@MenuFrame"))
				UserSettings.MenuFrame = getLink(item, UserSettings.MenuFrame);
			else if (ItemUtils.startsWithTag(item, "@DefaultFrame"))
				UserSettings.DefaultFrame = getLink(item,
						UserSettings.DefaultFrame);
			else if (ItemUtils.startsWithTag(item, "@AntiAlias"))
				UserSettings.AntiAlias = getBoolean(item,
						UserSettings.AntiAlias);
			else if (ItemUtils.startsWithTag(item, "@Gravity"))
				UserSettings.Gravity = getInt(item, UserSettings.Gravity);
			else if (ItemUtils.startsWithTag(item, "@ShowLineHighlight"))
				UserSettings.LineHighlight = getBoolean(item,
						UserSettings.LineHighlight);
			else if (ItemUtils.startsWithTag(item, "@LineStraightenThreshold"))
				UserSettings.LineStraightenThreshold = getInt(item,
						UserSettings.LineStraightenThreshold);
			else if (ItemUtils.startsWithTag(item, "@NoOpThreshold"))
				UserSettings.NoOpThreshold = getInt(item,
						UserSettings.NoOpThreshold);
			else if (ItemUtils.startsWithTag(item, "@InitialWidth")) {
				UserSettings.InitialWidth = getInt(item,
						UserSettings.InitialWidth);
			} else if (ItemUtils.startsWithTag(item, "@InitialHeight")) {
				UserSettings.InitialHeight = getInt(item,
						UserSettings.InitialHeight);
			} else if (ItemUtils.startsWithTag(item, "@TitlePosition")) {
				UserSettings.TitlePosition = getInt(item,
						UserSettings.TitlePosition);
			} else if (ItemUtils.startsWithTag(item, "@Logging")) {
				UserSettings.Logging = getBoolean(item, UserSettings.Logging);
			} else if (ItemUtils.startsWithTag(item,
					ItemUtils.TAG_ITEM_TEMPLATE)) {
				UserSettings.ItemTemplate = ((Text) item).getTemplateForm();
			} else if (ItemUtils.startsWithTag(item,
					ItemUtils.TAG_ANNOTATION_TEMPLATE)) {
				UserSettings.AnnotationTemplate = ((Text) item)
						.getTemplateForm();
			} else if (ItemUtils.startsWithTag(item,
					ItemUtils.TAG_STAT_TEMPLATE)) {
				UserSettings.StatTemplate = ((Text) item).getTemplateForm();
			} else if (ItemUtils.startsWithTag(item,
					ItemUtils.TAG_CODE_COMMENT_TEMPLATE)) {
				UserSettings.CodeCommentTemplate = ((Text) item)
						.getTemplateForm();
			} else if (ItemUtils.startsWithTag(item,
					ItemUtils.TAG_LINE_TEMPLATE)) {
				UserSettings.LineTemplate = ((Text) item).getTemplateForm();
			} else if (ItemUtils.startsWithTag(item, "@FramesetDir")) {
				String dir = getDir(item, null);
				if (dir != null)
					UserSettings.FrameDirs.add(dir);
			} else if (ItemUtils.startsWithTag(item, "@LogDir")) {
				org.expeditee.gui.FrameIO.LOGS_DIR = getDir(item,
						org.expeditee.gui.FrameIO.LOGS_DIR);
			} else if (ItemUtils.startsWithTag(item, "@ImageDir")) {
				String dir = getDir(item, null);
				if (dir != null)
					UserSettings.ImageDirs.add(0, dir);
			} else if (ItemUtils.startsWithTag(item, "@Threading")) {
				UserSettings.Threading = getBoolean(item,
						UserSettings.Threading);
			} else if (ItemUtils.startsWithTag(item, "@ColorWheel")) {
				Item.COLOR_WHEEL = getColorWheel(item);
			} else if (ItemUtils.startsWithTag(item, "@FillColorWheel")) {
				Item.FILL_COLOR_WHEEL = getColorWheel(item);
			} else if (ItemUtils.startsWithTag(item, "@BackgroundColorWheel")) {
				Frame.COLOR_WHEEL = getColorWheel(item);
			} else if (ItemUtils.startsWithTag(item, "@FramesetDirs")) {
				UserSettings.FrameDirs.addAll(getDirs(item));
			}
		}

		if (UserSettings.FirstFrame == null)
			UserSettings.FirstFrame = profile.getName();
		else {
			if (FrameIO.LoadFrame(UserSettings.FirstFrame) == null) {
				FrameGraphics.WarningMessage("Home frame: "
						+ UserSettings.FirstFrame + " could not be found.");
				UserSettings.FirstFrame = profile.getName();
			}
		}

	}

	private static Color[] getColorWheel(Item item) {
		String childName = item.getAbsoluteLink();
		if (childName != null) {
			Frame child = FrameIO.LoadFrame(childName);
			if (child != null) {
				List<Text> textItems = child.getBodyTextItems(true);
				Color[] colorList = new Color[textItems.size() + 1];
				for (int i = 0; i < textItems.size(); i++) {
					colorList[i] = textItems.get(i).getColor();
				}
				// Make the last item transparency or default for forecolor
				colorList[colorList.length - 1] = null;

				return colorList;
			}
		}
		return new Color[] { Color.black, Color.white, null};
	}

	private static String getLink(Item item, String alt) {
		if (item == null || !(item instanceof Text))
			return alt;

		String value = AttributeUtils.getValue(((Text) item).getFirstLine())
				.trim();
		if (value.length() > 0) {
			item.setLink(value);
			return value;
		} else if (item.getLink() != null)
			return item.getLink();

		return alt;
	}

	private static boolean getBoolean(Item item, boolean alt) {
		if (item == null || !(item instanceof Text))
			return alt;

		String value = AttributeUtils.getValue(((Text) item).getFirstLine())
				.trim().toLowerCase();
		if (value != null && value.length() > 0) {
			if (value.equals("t") || value.equals("true")
					|| value.equals("yes") || value.equals("y"))
				return true;

			if (value.equals("f") || value.equals("false")
					|| value.equals("no") || value.equals("n")) {
				return false;
			}
		}

		return alt;
	}

	private static int getInt(Item item, int alt) {
		if (item == null || !(item instanceof Text))
			return alt;

		String value = AttributeUtils.getValue(((Text) item).getFirstLine())
				.trim().toLowerCase();
		try {
			return Integer.parseInt(value);
		} catch (Exception e) {
		}

		return alt;
	}

	private static String getDir(Item item, String alt) {
		String name = getString(item, null);
		if (name != null) {
			File tester = new File(name);
			if (tester.exists() && tester.isDirectory()) {
				if (name.endsWith(File.separator))
					return name;
				else
					return name + File.separator;
			}
		}

		return alt;
	}

	private static ArrayList<String> getDirs(Item item) {
		ArrayList<String> dirsToAdd = new ArrayList<String>();
		String dirListFrameName = item.getAbsoluteLink();
		if (dirListFrameName != null) {
			Frame dirListFrame = FrameIO.LoadFrame(dirListFrameName);
			if (dirListFrame != null) {
				for (Text t : dirListFrame.getBodyTextItems(false)) {
					String dirName = t.getText().trim();
					File tester = new File(dirName);
					if (tester.exists() && tester.isDirectory()) {
						if (dirName.endsWith(File.separator))
							dirsToAdd.add(dirName);
						else
							dirsToAdd.add(dirName + File.separator);
					}
				}
			}
		}

		return dirsToAdd;
	}

	/*
	 * private static String FindFile(List<Item> items, String tag, String
	 * alt){ String name = FindString(items, tag, null); if(name != null){ File
	 * tester = new File(name); if(tester.exists()) return name; }
	 * 
	 * return alt; }
	 */
	private static String getString(Item item, String alt) {
		if (item == null || !(item instanceof Text))
			return alt;

		String value = AttributeUtils.getValue(((Text) item).getFirstLine())
				.trim();
		if (value != null && value.length() > 0)
			return value;

		return alt;
	}

	public static void Parse(Frame toParse) {
		Parse(toParse, false);
	}

	/**
	 * Checks for any special Annotation items and updates the display as
	 * necessary. Special Items: Images, overlays, sort.
	 * 
	 */
	public static void Parse(Frame toParse, boolean firstParse) {
		// System.out.println(firstParse);
		if (firstParse)
			ItemUtils.EnclosedCheck(toParse.getItems());
		List<Item> items = toParse.getItems();

		// if XRayMode is on, replace pictures with their underlying text
		if (FrameGraphics.isXRayMode()) {

			// BROOK: Must handle these a little different
			List<InteractiveWidget> widgets = toParse.getInteractiveWidgets();

			for (Item i : items) {
				if (i instanceof Picture) {
					toParse.removeItem(i);
					toParse.addItem(((Picture) i).getText());
				} else if (i instanceof WidgetCorner) {
					toParse.removeItem(i);
				} else if (i instanceof WidgetEdge) {
					toParse.removeItem(i);
				}
			}

			for (InteractiveWidget iw : widgets) {
				toParse.addItem(iw.getSource());
			}

			// update the list of items
			items.clear();
			items.addAll(toParse.getItems());
		}

		// Text title = null;
		// Text template = UserSettingsTemplate.copy();

		List<Overlay> overlays = new ArrayList<Overlay>();
		List<Vector> vectors = new ArrayList<Vector>();

		// disable reading of cached overlays if in twinframes mode
		if (DisplayIO.isTwinFramesOn())
			FrameIO.SuspendCache();

		int pointtype = Item.POINTTYPE_SQUARE;
		boolean filledPoints = true;

		int permissionLevel = getPermissionLevel(toParse.getProtection(),
				Item.PERMISSION_FULL);
		// check for any new overlay items
		for (Item i : items) {
			try {
				i.Permission = permissionLevel;
				if (ItemUtils.startsWithTag(i, ItemUtils.TAG_POINTTYPE)) {
					Text txt = (Text) i;
					String line = txt.getFirstLine();
					line = ItemUtils.StripTag(line, ItemUtils
							.GetTag(ItemUtils.TAG_POINTTYPE));

					if (line != null) {
						line = line.toLowerCase();
						if (line.indexOf(" ") > 0) {
							String fill = line.substring(line.indexOf(" ") + 1);
							if (fill.startsWith("nofill"))
								filledPoints = false;
							else
								filledPoints = true;
						}

						if (line.startsWith("circle"))
							pointtype = Item.POINTTYPE_CIRCLE;
						else
							pointtype = Item.POINTTYPE_SQUARE;
					}
				}

				// check for Images and widgets
				if (!FrameGraphics.isXRayMode()) {
					if (ItemUtils.startsWithTag(i, ItemUtils.TAG_IMAGE))
						createPicture(toParse, (Text) i);
					// check for frame images
					else if (i.getLink() != null
							&& !i.getAbsoluteLink().equalsIgnoreCase(
									toParse.getName())) {
						if (ItemUtils.startsWithTag(i,
								ItemUtils.TAG_FRAME_IMAGE)) {
							createFramePicture(toParse, (Text) i);
						} else if (ItemUtils.startsWithTag(i,
								ItemUtils.TAG_BITMAP_IMAGE)) {
							createFrameBitmap(toParse, (Text) i);
						}
					}
					// Check for interactive widgets
					else if (ItemUtils.startsWithTag(i, ItemUtils.TAG_IWIDGET)) {
						createWidget(toParse, (Text) i);
					}
				}

				// check for new VECTOR items
				if (ItemUtils.startsWithTag(i, ItemUtils.TAG_VECTOR)
						&& i.getLink() != null) {
					Frame vector = FrameIO.LoadFrame(i.getAbsoluteLink());
					if (vector != null) {
						String scaleString = AttributeUtils.getValue(((Text) i)
								.getText());
						Float scale = 1F;
						try {
							scale = Float.parseFloat(scaleString);
						} catch (Exception e) {
						}
						vectors.add(new Vector(vector, i.getPosition(), scale));
					}
				}

				// check for new OVERLAY items
				if (ItemUtils.startsWithTag(i, ItemUtils.TAG_OVERLAY)
						&& i.getLink() != null) {
					Frame overlay = FrameIO.LoadFrame(i.getAbsoluteLink());
					// Parse(overlay);
					if (overlay != null && !overlays.contains(overlay))
						overlays
								.add(new Overlay(overlay, Item.PERMISSION_NONE));
				}

				// check for ACTIVE_OVERLAY items
				if (ItemUtils.startsWithTag(i, ItemUtils.TAG_ACTIVE_OVERLAY)
						&& i.getLink() != null) {
					String link = i.getAbsoluteLink();
					Frame overlay = FrameIO.LoadFrame(link);
					// Parse(overlay);
					// get level if specified
					String level = AttributeUtils.getValue(((Text) i)
							.getFirstLine());
					// default permission (if none is specified)
					int permission = getPermissionLevel(level,
							Item.PERMISSION_FOLLOW_LINKS);

					if (overlay != null) {
						// Add it if its a new overlay
						boolean found = false;
						// Check if this overlay is already in the list and
						// adjust its permissions
						for (Overlay o : overlays) {
							if (o.Frame.equals(overlay)) {
								o.Level = permission;
								found = true;
								break;
							}
						}
						// If it wasnt in the list create it and add it.
						if (!found) {
							overlays.add(new Overlay(overlay, permission));
						}
					}
				}

				/*
				 * check for Item templates if (ItemUtils.isTag(i,
				 * ItemUtils.TAG_ITEM_TEMPLATE)) { template = (Text) i; }
				 */

				// only remove items on first parse
				if (firstParse) {
					// remove dots of size 1
					if (i instanceof Dot) {
						Item dot = (Item) i;
						if (dot.getSize() <= 1
								&& (dot.getLines() == null || dot.getLines()
										.size() == 0))
							if (dot.getConstraints() == null
									|| dot.getConstraints().size() == 0) {
								toParse.removeItem(dot);
							}
					}

					// remove Text that is all whitespace
					if (i instanceof Text) {
						List<String> text = ((Text) i).getTextList();
						// remove empty text items
						if (text == null || text.size() == 0)
							toParse.removeItem(i);
						else {
							for (String s : text)
								if (s.trim().length() > 0)
									break;
								else
									toParse.removeItem(i);
						}
					}

				}

			} catch (Exception e) {
				Logger.Log(e);
				e.printStackTrace();
				FrameGraphics.WarningMessage("Exception occured when loading "
						+ i.getClass().getSimpleName() + "(ID: " + i.getID()
						+ ") " + e.getMessage());
			}
		}

		for (Item i : items) {
			if (i instanceof Dot) {
				((Dot) i).setPointType(pointtype);
				((Dot) i).useFilledPoints(filledPoints);
			}
		}

		FrameIO.ResumeCache();

		synchronized (toParse) {
			toParse.clearOverlays();
			toParse.addAllOverlays(overlays);
			toParse.clearVectors();
			toParse.addAllVectors(vectors);
		}
	}

	/**
	 * Converts a string containing the permission level into its integer
	 * equivalent.
	 * 
	 * @param level
	 * @param defaultPermission
	 * @return
	 */
	public static int getPermissionLevel(String level, int defaultPermission) {
		level = level.trim().toLowerCase();
		if (level.length() == 0)
			return defaultPermission;

		try {
			return Integer.parseInt(level);
		} catch (Exception e) {
			// if this is a text permission level, parse it

			if (level.equals("none"))
				return Item.PERMISSION_NONE;
			else if (level.equals("links") || level.equals("link"))
				return Item.PERMISSION_FOLLOW_LINKS;
			else if (level.equals("copy"))
				return Item.PERMISSION_COPY;
			else if (level.equals("tdfc"))
				return Item.PERMISSION_TDFC;
			else if (level.equals("full") || level.equals("all"))
				return Item.PERMISSION_FULL;
		}
		return defaultPermission;
	}

	/**
	 * Searches through the list of items on this frame to find one at the given
	 * x,y coordinates.
	 * 
	 * @param x
	 *            The x coordinate
	 * @param y
	 *            The y coordinate
	 * @return The Item at the given coordinates, or NULL if none is found.
	 */
	public static Item onItem(Frame toCheck, int x, int y) {
		if (toCheck == null)
			return null;

		List<Item> possibles = new ArrayList<Item>(0);

		// if the mouse is in the message area
		if (y > FrameGraphics.getMaxFrameSize().getHeight()) {
			// check the individual message items
			for (Item message : FrameGraphics.Messages) {
				if (message != null) {
					if (message.contains(x, y)) {
						message.Permission = Item.PERMISSION_COPY;
						possibles.add(message);
					} else
						// Not sure why but if the line below is removed then
						// several items can be highlighted at once
						message.setSelectedMode(Item.SelectedMode.None);
				}
			}

			// check the link to the message frame
			if (FrameGraphics.MessageLink != null) {
				if (FrameGraphics.MessageLink.contains(x, y)) {
					FrameGraphics.MessageLink.Permission = Item.PERMISSION_COPY;
					possibles.add(FrameGraphics.MessageLink);
				}
			}

			// this is taken into account in contains
			// y -= FrameGraphics.getMaxFrameSize().height;
			// otherwise, the mouse is on the frame
		} else {

			if (LastEdited != null && LastEdited.contains(x, y)
					&& !Frame.FreeItems.contains(LastEdited)
					&& LastEdited.getParent() == DisplayIO.getCurrentFrame()
					&& LastEdited.getParent().getItems().contains(LastEdited)) {
				LastEdited.Permission = Item.PERMISSION_FULL;
				return LastEdited;
			} else
				LastEdited = null;

			ArrayList<Item> checkList = new ArrayList<Item>();
			checkList.addAll(toCheck.getItems());
			checkList.add(toCheck.getNameItem());
			for (Item i : checkList) {
				// do not check annotation items in audience mode
				if (!(FrameGraphics.isAudienceMode() && i.isAnnotation())) {
					if (i.contains(x, y) && !Frame.FreeItems.contains(i)) {
						// names have copy permissions only
						if (i == toCheck.getNameItem()) {
							// i.Permission = Item.PERMISSION_TDFC;
							possibles.add(i);
						} else {
							// i.Permission = Item.PERMISSION_FULL;
							possibles.add(i);
						}
					}
				}
			}
			int framePermission = getPermissionLevel(toCheck.getProtection(),
					Item.PERMISSION_FULL);
			for (Overlay o : toCheck.getOverlays()) {
				if (o.Level > Item.PERMISSION_NONE) {
					Item i = onItem(o.Frame, x, y);

					// the message is included above with full permissions,
					// it should be ignored here
					if (i != null && i.getID() > 0) {
						i.Permission = Math.min(o.Level, framePermission);
						possibles.add(i);
					}
				}
			}
		}

		// if there are no possible items, return null
		if (possibles.size() == 0)
			return null;

		// if there is only one possibility, return it
		if (possibles.size() == 1)
			return possibles.get(0);

		// return closest x,y pair to mouse
		Item closest = possibles.get(0);
		int distance = (int) Math.round(Math.sqrt(Math.pow(Math.abs(closest
				.getX()
				- x), 2)
				+ Math.pow(Math.abs(closest.getY() - y), 2)));

		for (Item i : possibles) {

			int d = (int) Math.round(Math.sqrt(Math.pow(Math.abs(i.getX() - x),
					2)
					+ Math.pow(Math.abs(i.getY() - y), 2)));

			// System.out.println(d);
			if (d <= distance) {
				distance = d;

				// dots take precedence over lines
				if ((!(closest instanceof Dot && i instanceof Line))
						&& (!(closest instanceof Text && i instanceof Line)))
					closest = i;

			}

		}

		return closest;
	}

	public static Item getCurrentItem() {
		return onItem(DisplayIO.getCurrentFrame(), DisplayIO.getMouseX(),
				DisplayIO.getMouseY());
	}

	public static List<Item> getCurrentItems() {

		Collection<Item> enclosure = getEnclosingDots();
		if (enclosure == null || enclosure.size() == 0)
			return null;

		Item firstItem = enclosure.iterator().next();

		List<Item> enclosed = getItemsEnclosedBy(DisplayIO.getCurrentFrame(),
				firstItem.getEnclosedShape());
		for (Item i : enclosed) {
			if (!enclosure.contains(i))
				i.setSelectedMode(Item.SelectedMode.None);
		}

		return enclosed;
	}

	public static Collection<Item> getEnclosingDots() {
		// update enclosed shapes
		Frame current = DisplayIO.getCurrentFrame();
		List<Item> items = current.getItems();

		// Remove all items that are connected to freeItems
		List<Item> freeItems = new ArrayList<Item>(Frame.FreeItems);
		while (freeItems.size() > 0) {
			Item item = freeItems.get(0);
			Collection<Item> connected = item.getAllConnected();
			items.removeAll(connected);
			freeItems.removeAll(connected);
		}

		List<Item> used = new ArrayList<Item>(0);

		while (items.size() > 0) {
			Item i = items.get(0);
			items.remove(i);
			if (i.isLineEnd() && i.isEnclosed()) {
				Polygon p = i.getEnclosedShape();
				if (p.contains(DisplayIO.getMouseX(), DisplayIO.getMouseY())) {
					used.add(i);
					items.removeAll(i.getEnclosingDots());
				}
			}
		}

		if (used.size() == 0)
			return null;

		// if there is only one possibility, return it
		if (used.size() == 1) {
			return used.get(0).getEnclosingDots();
			// otherwise, determine which polygon is closest to the cursor
		} else {
			Collections.sort(used, new Comparator<Item>() {
				public int compare(Item d1, Item d2) {
					Polygon p1 = d1.getEnclosedShape();
					Polygon p2 = d2.getEnclosedShape();

					int closest = Integer.MAX_VALUE;
					int close2 = Integer.MAX_VALUE;

					int mouseX = DisplayIO.getMouseX();
					int mouseY = DisplayIO.getMouseY();

					for (int i = 0; i < p1.npoints; i++) {
						int diff = Math.abs(p1.xpoints[i] - mouseX)
								+ Math.abs(p1.ypoints[i] - mouseY);
						int diff2 = Integer.MAX_VALUE;

						if (i < p2.npoints)
							diff2 = Math.abs(p2.xpoints[i] - mouseX)
									+ Math.abs(p2.ypoints[i] - mouseY);

						if (diff < Math.abs(closest)) {
							close2 = closest;
							closest = diff;
						} else if (diff < Math.abs(close2))
							close2 = diff;

						if (diff2 < Math.abs(closest)) {
							close2 = closest;
							closest = -diff2;
						} else if (diff2 < Math.abs(close2))
							close2 = diff2;
					}

					if (closest > 0 && close2 > 0)
						return -10;

					if (closest < 0 && close2 < 0)
						return 10;

					if (closest > 0)
						return -10;

					return 10;
				}

			});

			return used.get(0).getEnclosingDots();
		}
	}

	public static List<Item> getItemsEnclosedBy(Frame frame, Polygon poly) {
		List<Item> contained = frame.getItemsWithin(poly);

		ArrayList<Item> results = new ArrayList<Item>(contained.size());

		// check for correct permissions
		for (Item item : contained) {
			// if the item is on the frame
			if (item.getParent() == frame || item.getParent() == null) {
				// item.Permission = Item.PERMISSION_FULL;
				results.add(item);
				// otherwise, it must be on an overlay frame
			} else {
				for (Overlay overlay : frame.getOverlays())
					if (overlay.Frame == item.getParent()) {
						item.Permission = overlay.Level;
						results.add(item);
						break;
					}
			}

		}

		return results;
	}

	/**
	 * Fills the given Frame with default profile tags
	 */
	public static void CreateDefaultProfile(Frame profile) {
		Text title = profile.getTitleItem();
		title.setText("Profile Frame");
		title.setSize(50);
		title.setFontStyle("Bold");
		title.setFamily("SansSerif");
		title.setColor(Color.BLUE);
		title.setPosition(25, 50);

		int spacing = 50;
		final int intialYPos = 75;
		int xPos = 75;
		int yPos = intialYPos;

		yPos += spacing;
		profile.addText(xPos, yPos, "@HomeFrame: " + profile.getName(), null,
				profile.getName());
		yPos += spacing;
		String defaultFrameName = profile.getFramesetName() + "0";
		profile.addText(xPos, yPos, "@DefaultFrame: " + defaultFrameName, null,
				defaultFrameName);
		yPos += spacing;

		profile.addText(xPos, yPos, "@InitialWidth: "
				+ UserSettings.InitialWidth, null);
		yPos += spacing;

		profile.addText(xPos, yPos, "@InitialHeight: "
				+ UserSettings.InitialHeight, null);
		yPos += spacing;

		Text t = profile.addText(xPos, yPos, "@ItemTemplate", null);
		t.setColor(Color.black);

		yPos += spacing;
		t = profile.addText(xPos, yPos, "@AnnotationTemplate", null);
		t.setColor(Color.magenta);

		yPos += spacing;
		t = profile.addText(xPos, yPos, "@CommentTemplate", null);
		t.setColor(Item.COLOR_WHEEL[3]);

		yPos += spacing;
		t = profile.addText(xPos, yPos, "@StatTemplate", null);
		t.setColor(Color.BLACK);
		t.setBackgroundColor(new Color(0.9F, 0.9F, 0.9F));
		t.setFamily(Text.MONOSPACED_FONT);
		t.setSize(14);

		xPos = 600;
		yPos = intialYPos;

		// Iterate through the help folder and add links
		File helpDirectory = new File(FrameIO.HELP_PATH);
		if (helpDirectory != null) {
			File[] helpFramesets = helpDirectory.listFiles();
			if (helpFramesets != null) {

				// Add the title for the help index
				Text help = profile.addText(xPos, yPos, "Expeditee Help", null);
				help.setSize(25);
				help.setFontStyle("Bold");
				help.setFamily("SansSerif");
				help.setColor(Item.COLOR_WHEEL[3]);

				xPos = 625;

				for (File helpFrameset : helpFramesets) {
					String framesetName = helpFrameset.getName();
					Frame indexFrame = FrameIO.LoadFrame(framesetName + '1');
					// Look through the folder for help index pages
					if (indexFrame != null
							&& ItemUtils.FindTag(indexFrame.getItems(),
									"@HelpIndex") != null) {
						yPos += spacing;
						t = profile.addText(xPos, yPos, indexFrame
								.getFramesetName(), null);
						t.setLink(indexFrame.getName());
					}
				}
			}
		}

		// FrameUtils.Parse(profile);
		FrameIO.SaveFrame(profile);
	}

	private static void checkTDFCItemWaiting(Frame currentFrame) {
		Item tdfcItem = FrameUtils.getTdfcItem();
		// if there is a TDFC Item waiting
		if (tdfcItem != null) {
			boolean change = currentFrame.hasChanged();
			boolean saved = currentFrame.isSaved();
			// Save the parent of the item if it has not been saved
			if (!change && !saved) {
				tdfcItem.setLink(null);
				tdfcItem.getParent().setChanged(true);
				FrameIO.SaveFrame(tdfcItem.getParent());
				FrameGraphics.Repaint();
			} else {
				SessionStats.CreatedFrame();
			}

			setTdfcItem(null);
		}
	}

	public static void setTdfcItem(Item _tdfcItem) {
		FrameUtils._tdfcItem = _tdfcItem;
	}

	public static Item getTdfcItem() {
		return FrameUtils._tdfcItem;
	}

	private static Item _tdfcItem = null;
}
